#include <iostream>
#include <vector>
#include <algorithm>
#include <map>
#include <time.h>
#include <assert.h>
#include <fstream>

using namespace std;

struct indx {
    int i,j;
};


indx operator-(const indx &p1, const indx &p2) {
    indx out={p1.i-p2.i,p1.j-p2.j};
    return out;
}

indx operator+(const indx &p1, const indx &p2) {
    indx out={p1.i+p2.i,p1.j+p2.j};
    return out;
}

bool operator <(const indx& x, const indx& y) {
    return (x.i==y.i)?x.j<y.j:x.i<y.i;
}

bool operator ==(const indx& x, const indx& y) {
    return (x.i==y.i)&&(x.j==y.j);
}

struct cell {
    indx place;
    char p;
    bool is_stem;
};

//global variables
char **lattice;
int N1=50, N2=50; //lattice dimension
int shiftH=0, shiftV=0;
int gN=50; //how many rows/cols add when expanding

int cc=0,csc=1;
vector<pair<int, int> > trajectory;
vector<pair<int, int> > trajectoryLattice;


vector<cell> cells; //vector containing all cells present in the system
vector<indx> directions;
map<indx, vector<indx > > append_direction;

//simualtion specific parmeter
char pmax=10; //divisions before proliferative capacity exhaustion
double pDiv=1./24.; //division probability
double alpha=0.01; //spontaneous death probability
double ps=0.1; //probability of symmetric division
double pmig=10./24.; //probability of migration

void setPlaceVal(indx place, char val) {
    lattice[place.i+shiftV][place.j+shiftH]=val;
}

void plotLattice() {
    for (int i=0; i<N1; i++) {
        cout << endl;
        for (int j=0; j<N2; j++)
            cout << (int)lattice[i][j] << " ";
    }
    cout << endl;
}



char getLatticeValS(indx place) {
    return lattice[place.i+shiftV][place.j+shiftH];
}


char getPlaceVal(indx place) {
    
    place.i+=shiftV;
    place.j+=shiftH;

    if (place.i<0) {//expand lattice top
        char **aux;
        aux=new char *[N1+gN];
        
        for (int i=0; i<gN; i++) {
            aux[i] = new char [N2] ();
            fill_n(aux[i], N2, 9);
        }
        
        for(int i=gN; i<N1+gN; i++) {
            aux[i]=lattice[i-gN];
        }
        delete [] lattice;
        lattice=aux;
        
        shiftV=shiftV+gN;
        place.i+=gN;
        N1=N1+gN;

    }
    
    if (place.j<0) {//expand lattice left
        for (int i=0; i<N1; i++) {
            char *aux;
            aux = new char [N2+gN] ();
            fill_n(aux, gN, 9);
            memcpy(aux+gN, lattice[i],N2*sizeof(char));
            delete [] lattice[i];
            lattice[i] = aux;
        }
        
        shiftH=shiftH+gN;
        place.j+=gN;
        N2=N2+gN;

    }
    
    if (place.i>=N1) {//expand lattice bottom
        char **aux;
        aux=new char *[N1+gN];
        
        for(int i=0; i<N1; i++) {
            aux[i]=lattice[i];
        }
        
        for (int i=N1; i<gN+N1; i++) {
            aux[i] = new char [N2] ();
            fill_n(aux[i], N2, 9);
        }
        
        delete [] lattice;
        lattice=aux;
        
        N1=N1+gN;
    }
    
    if (place.j>=N2) {//expand lattice right
        for (int i=0; i<N1; i++) {
            char *aux;
            aux = new char [N2+gN] ();
            memcpy(aux, lattice[i],N2*sizeof(char));
            fill_n(aux+N2, gN, 9);
            delete [] lattice[i];
            lattice[i] = aux;
        }
        
        N2=N2+gN;
    }
    
    return lattice[place.i][place.j];
}


indx returnEmptyPlace (indx place) {//FUNCTION RETTURNING RANDOM FREE SPOT (IF THERE IS ANY)
    if(lattice[place.i+shiftV][place.j+shiftH]>0) {
        indx aux;
        random_shuffle(directions.begin(), directions.end());
        for(int i=0; i<8; i++) {
            aux.i=directions.at(i).i+place.i; aux.j=directions.at(i).j+place.j;
            if (getPlaceVal(aux)==9)
                return aux;
        }
    }
    return place;
    
}

/////////////// update the number of free spots after cell proliferation
void update_proliferation(indx place) {
	setPlaceVal(place,8);
    for(int i=0; i<8; i++)
        if (getPlaceVal(place+directions.at(i))<9) {
            lattice[place.i+shiftV+directions.at(i).i][place.j+shiftH+directions.at(i).j]--;
            lattice[place.i+shiftV][place.j+shiftH]--;
        }
}
///////////////

/////////////// update the number of free spots after cell death
void update_death(indx place) {
	for(int i=0; i<8; i++)
		if (getPlaceVal(place+directions.at(i))<9) {
			lattice[place.i+shiftV+directions.at(i).i][place.j+shiftH+directions.at(i).j]++;
		}
}
///////////////

/////////////// update the number of free spots after cell migration
void update_migration(indx prev, indx next, char prevFreeSpots) {
	vector<indx> dir_add(append_direction[next-prev]);
	vector<indx> dir_sub(append_direction[prev-next]);
	unsigned char pom1=0,pom2=0;
    
    
	for (vector<indx>::iterator it=dir_sub.begin();it!=dir_sub.end(); it++)
		if (getPlaceVal(next+(*it))<9 ) {
			lattice[next.i+(*it).i+shiftV][next.j+(*it).j+shiftH]--;
			pom1++;
		}
    
	for (vector<indx>::iterator it=dir_add.begin();it!=dir_add.end(); it++)
		if (getPlaceVal(prev+(*it))<9) {
			lattice[prev.i+(*it).i+shiftV][prev.j+(*it).j+shiftH]++;
			pom2++;
		}
	lattice[next.i+shiftV][next.j+shiftH]=prevFreeSpots-(dir_add.size()-pom2)+(dir_sub.size()-pom1);
}
///////////////

vector<indx> wall(short x, short y) {
	vector<indx> out;
	short imin=-1,imax=1,jmin=-1,jmax=1;
	if (x==-1) {imin=1;}
	else if (x==1) {imax=-1;}
	else if (y==1) {jmax=-1;}
	else if (y==-1) {jmin=1;}
	
    indx aux;
    
	for (short i=imin; i<=imax; i++ )
        for (short j=jmin; j<=jmax; j++ ) {
            aux.i=i; aux.j=j;
            out.push_back(aux);
        }

	return out;
}

vector<indx> corner(short x, short y) {
	vector<indx> out,out2,pom;
	vector<indx>::iterator it,it2;
    
	if (x!=0) {pom=wall(x,0); out.insert(out.end(),pom.begin(),pom.end());}
	if (y!=0) {pom=wall(0,y); out.insert(out.end(),pom.begin(),pom.end());}
	
	for (it=out.begin(); it!=out.end(); it++) {
		it2=find(out2.begin(), out2.end(), (*it));
		if (it2==out2.end())
            out2.push_back((*it));
		
	}
    
	return out2;
}


void initialize() {//FUNCTON INITIALIZING CA
    lattice=new char *[N1];
    for (int i=0; i<N1; i++) {
        lattice[i] = new char [N2] ();
        fill_n(lattice[i], N2, 9);
    }
    
    lattice[N1/2][N2/2]=8;
    indx primaryPlace={N1/2, N2/2};
    cell primaryCell={primaryPlace, pmax, true};
    cells.push_back(primaryCell);
    
    indx aux;
    for(int i=-1; i<2; i++)
        for(int j=-1; j<2;j++)
            if (j!=0 || i!=0) {
                aux.i=i; aux.j=j;
                directions.push_back(aux);
                append_direction[aux]=corner(i,j);
            }
    
    aux.i=0; aux.j=0;
	append_direction[aux]=directions;
    
}

void clearCA(){
    for (int i=0; i<N1; i++) {
        delete [] lattice[i];
    }
    delete[] lattice;
}

void save_structure(int number){
	
	
	char tmp[50],tmp2[50];
	sprintf(tmp, "_%i.csv", number);
	strcat(tmp2,"struct");
    strcat(tmp2,tmp);
    ofstream file_op(tmp2);
    
    
	if (file_op) {
        file_op<<"x coord, y coord, is_stem, p, nFreeSpots"<<endl;
        vector<cell>::iterator it;
        
        file_op<< N1 <<", " << N2 << ", 0, 0, 0";
        
        for (it=cells.begin(); it!=cells.end(); it++) {
            file_op<<endl;
            file_op<< it->place.i+shiftV << ", " << it->place.j+shiftH << ", "<< it->is_stem <<", "<<(int)it->p<<", "<< (int)getLatticeValS(it->place);
        }
        file_op.close();
	}
}

void simulate(int nSteps) {//MAIN SIMUALTION PROCEDURE
    
    vector<cell> cellsTmp;
    double fate;
    cell currCell;
    indx newPlace;
    
    for (int i=0; i<nSteps; i++) {
       
        if ((i % 24) == 0 ) {
            //cout << "Day no.: " << i/24 << ", populationSize: " << cells.size() << " domain size: "<< N1 << "x" << N2 << endl;
            //cout << cells.size() << " ";
            pair<int, int> currSt (cc,csc);
            trajectory.push_back(currSt);
            pair<int, int> lattSize (N1,N2);
            trajectoryLattice.push_back(lattSize);
        }
        
        if ((i % (24*30)) == 0 ) {
            save_structure(i);
        }
        
        random_shuffle(cells.begin(), cells.end());
        
        while (!cells.empty()) {
            
           currCell=cells.back(); //pick the cell
           cells.pop_back();
           
            fate=((double)rand()/(double)RAND_MAX);
           
            if (fate<alpha && !currCell.is_stem) { //spontaneous death occurs
                setPlaceVal(currCell.place,9);
                update_death(currCell.place);
                cc--;
                //cout << "death" << endl;
            } else {

                newPlace = returnEmptyPlace(currCell.place);

                if (newPlace==currCell.place) {//no available space around
                    cellsTmp.push_back(currCell);
                } else {
                    fate=((double)rand()/(double)RAND_MAX);
                    if (fate<pDiv) {//division
                        //cout << "division" << endl;
                        cell newCell=currCell;
                        newCell.place=newPlace;
                        if (currCell.is_stem) {//if stem cell is divding
                            fate=((double)rand()/(double)RAND_MAX);
                            if (fate>ps) {//asymmetric division
                                newCell.is_stem=false;
                                cc++;
                            } else {
                                csc++;
                            }
                            cellsTmp.push_back(currCell);
                            cellsTmp.push_back(newCell);
                            update_proliferation(newPlace);
                        } else if (currCell.p==0) {//capacity exhausted
                            //cout << "capacity ex" << endl;
                            setPlaceVal(currCell.place,9);
                            update_death(currCell.place);
                            cc--;
                        } else {
                            newCell.p-=1; currCell.p-=1;
                            cellsTmp.push_back(currCell);
                            cellsTmp.push_back(newCell);
                            update_proliferation(newPlace);
                            cc++;
                        }
        
                    } else if (fate<pmig) {//migration
                        //cout << "migration" << endl;
                        update_migration(currCell.place,newPlace,getLatticeValS(currCell.place));
                        setPlaceVal(currCell.place,9);
                        currCell.place=newPlace;
                        cellsTmp.push_back(currCell);
                    } else { //do nothing
                        cellsTmp.push_back(currCell);
                    }
                }
             
            }
     
        }

        cells.swap(cellsTmp);
    }
}



int main() {
    clock_t start, stop;
    double t = 0.0;
    srand(time(NULL)); //initialize random number generator
    
    initialize(); //initialize CA
    //cout << (int)lattice[1][0] << endl;
    
    
    /* Start timer */
    assert((start = clock())!=-1);
    
    simulate(24*365);
    
    stop = clock();
    t = (double) (stop-start)/CLOCKS_PER_SEC;
    //cout << endl << "Run time: " << t << endl;
   
    ofstream outfile;
    
    outfile.open("outData.txt", std::ios_base::app);
    for (int i=0; i<trajectory.size(); i++)
        outfile << trajectory.at(i).first << " ";
    
    outfile << endl;
    
    for (int i=0; i<trajectory.size(); i++)
        outfile << trajectory.at(i).second << " ";
    
    outfile << endl;
    outfile.close();
    
    outfile.open("outDataLattice.txt", std::ios_base::app);
    for (int i=0; i<trajectoryLattice.size(); i++)
        outfile << trajectoryLattice.at(i).first << " ";
    
    outfile << endl;
    
    for (int i=0; i<trajectoryLattice.size(); i++)
        outfile << trajectoryLattice.at(i).second << " ";
    
    outfile << endl;
    outfile.close();
    
    outfile.open("outRunTimes.txt", std::ios_base::app);

    outfile << t << endl;

    outfile.close();

    save_structure(1);
    
    clearCA();
   
    cout << "done!" << endl;
    return 0;
}

